@@ -29,15 +29,23 @@ from hypothesis.database import ExampleDatabase, InMemoryExampleDatabase
 from hypothesis.errors import FailedHealthCheck, FlakyStrategyDefinition
 from hypothesis.internal.compat import PYPY, bit_count, int_from_bytes
 from hypothesis.internal.conjecture import engine as engine_module
-from hypothesis.internal.conjecture.data import ConjectureData, IRNode, Overrun, Status
+from hypothesis.internal.conjecture.data import (
+    ConjectureData,
+    IRNode,
+    Overrun,
+    Status,
+    ir_size_nodes,
+)
 from hypothesis.internal.conjecture.datatree import compute_max_children
 from hypothesis.internal.conjecture.engine import (
+    BUFFER_SIZE_IR,
     MIN_TEST_CALLS,
     ConjectureRunner,
     ExitReason,
     HealthCheckState,
     RunIsComplete,
 )
+from hypothesis.internal.conjecture.junkdrawer import startswith
 from hypothesis.internal.conjecture.pareto import DominanceRelation, dominance
 from hypothesis.internal.conjecture.shrinker import Shrinker
 from hypothesis.internal.entropy import deterministic_PRNG
@@ -49,45 +57,32 @@ from tests.conjecture.common import (
     SOME_LABEL,
     TEST_SETTINGS,
     buffer_size_limit,
+    ir,
     ir_nodes,
-    run_to_buffer,
+    run_to_nodes,
     shrinking_from,
 )
 
 
-def test_can_index_results():
-    @run_to_buffer
-    def f(data):
-        data.draw_bytes(5, 5)
-        data.mark_interesting()
-
-    assert f.index(0) == 0
-    assert f.count(0) == 5
-
-
 def test_non_cloneable_intervals():
-    @run_to_buffer
-    def x(data):
+    @run_to_nodes
+    def nodes(data):
         data.draw_bytes(10, 10)
         data.draw_bytes(9, 9)
         data.mark_interesting()
 
-    assert x == bytes(19)
+    assert tuple(n.value for n in nodes) == (bytes(10), bytes(9))
 
 
 def test_deletable_draws():
-    @run_to_buffer
-    def x(data):
+    @run_to_nodes
+    def nodes(data):
         while True:
             x = data.draw_bytes(2, 2)
             if x[0] == 255:
                 data.mark_interesting()
 
-    assert x == bytes([255, 0])
-
-
-def zero_dist(random, n):
-    return bytes(n)
+    assert tuple(n.value for n in nodes) == (b"\xff\x00",)
 
 
 def test_can_load_data_from_a_corpus():
@@ -122,7 +117,7 @@ def test_terminates_shrinks(n, monkeypatch):
     db = InMemoryExampleDatabase()
 
     def generate_new_examples(self):
-        self.cached_test_function([255] * 1000)
+        self.cached_test_function_ir(ir(255) * 1000)
 
     monkeypatch.setattr(
         ConjectureRunner, "generate_new_examples", generate_new_examples
@@ -144,20 +139,21 @@ def test_terminates_shrinks(n, monkeypatch):
 
 
 def test_detects_flakiness():
-    failed_once = [False]
-    count = [0]
+    failed_once = False
+    count = 0
 
     def tf(data):
+        nonlocal count, failed_once
         data.draw_bytes(1, 1)
-        count[0] += 1
-        if not failed_once[0]:
-            failed_once[0] = True
+        count += 1
+        if not failed_once:
+            failed_once = True
             data.mark_interesting()
 
     runner = ConjectureRunner(tf)
     runner.run()
     assert runner.exit_reason == ExitReason.flaky
-    assert count == [MIN_TEST_CALLS + 1]
+    assert count == MIN_TEST_CALLS + 1
 
 
 def recur(i, data):
@@ -184,7 +180,7 @@ def test_variadic_draw():
         result = []
         while True:
             data.start_example(SOME_LABEL)
-            d = data.draw_bytes(1, 1)[0] & 7
+            d = data.draw_integer(0, 2**8 - 1) & 7
             if d:
                 result.append(data.draw_bytes(d, d))
             data.stop_example()
@@ -192,25 +188,32 @@ def test_variadic_draw():
                 break
         return result
 
-    @run_to_buffer
-    def b(data):
+    @run_to_nodes
+    def nodes(data):
         if any(all(d) for d in draw_list(data)):
             data.mark_interesting()
 
-    ls = draw_list(ConjectureData.for_buffer(b))
+    ls = draw_list(ConjectureData.for_ir_tree(nodes))
     assert len(ls) == 1
     assert len(ls[0]) == 1
 
 
-def test_draw_to_overrun():
-    @run_to_buffer
-    def x(data):
+def test_draw_to_overrun(monkeypatch):
+    # TODO_BETTER_SHRINK: sometimes we can get unlucky and fail to shrink the
+    # initial size draw d to 2 before shrinking the 128 * d bytes, but I'm not
+    # sure why.
+    #
+    # If we do get unlucky in such a way then we need more than 500 shrinks to finish.
+    monkeypatch.setattr(engine_module, "MAX_SHRINKS", 1000)
+
+    @run_to_nodes
+    def nodes(data):
         d = (data.draw_bytes(1, 1)[0] - 8) & 0xFF
         data.draw_bytes(128 * d, 128 * d)
         if d >= 2:
             data.mark_interesting()
 
-    assert x == bytes([10]) + bytes(128 * 2)
+    assert tuple(n.value for n in nodes) == (bytes([10]),) + (bytes(128 * 2),)
 
 
 def test_can_navigate_to_a_valid_example():
@@ -258,15 +261,17 @@ def test_stops_after_max_examples_when_generating():
 @pytest.mark.parametrize("examples", [1, 5, 20, 50])
 def test_stops_after_max_examples_when_generating_more_bugs(examples):
     seen = []
-    bad = [False, False]
+    err_common = False
+    err_rare = False
 
     def f(data):
         seen.append(data.draw_integer(0, 2**32 - 1))
         # Rare, potentially multi-error conditions
+        nonlocal err_common, err_rare
         if seen[-1] > 2**31:
-            bad[0] = True
+            err_rare = True
             raise ValueError
-        bad[1] = True
+        err_common = True
         raise Exception
 
     runner = ConjectureRunner(
@@ -278,14 +283,14 @@ def test_stops_after_max_examples_when_generating_more_bugs(examples):
         pass
     # No matter what, whether examples is larger or smalller than MAX_TEST_CALLS,
     # we stop looking at max_examples.  (and re-run each failure for the traceback)
-    assert len(seen) <= examples + sum(bad)
+    assert len(seen) <= examples + err_common + err_rare
 
 
 def test_interleaving_engines():
     children = []
 
-    @run_to_buffer
-    def x(data):
+    @run_to_nodes
+    def nodes(data):
         rnd = Random(data.draw_bytes(1, 1))
 
         def g(d2):
@@ -298,7 +303,7 @@ def test_interleaving_engines():
         if runner.interesting_examples:
             data.mark_interesting()
 
-    assert x == b"\0"
+    assert tuple(n.value for n in nodes) == (b"\0",)
     for c in children:
         assert not c.interesting_examples
 
@@ -337,38 +342,40 @@ def test_reuse_phase_runs_for_max_examples_if_generation_is_disabled():
 
 
 def test_erratic_draws():
-    n = [0]
+    n = 0
 
     with pytest.raises(FlakyStrategyDefinition):
 
-        @run_to_buffer
-        def x(data):
-            data.draw_bytes(n[0], n[0])
-            data.draw_bytes(255 - n[0], 255 - n[0])
-            if n[0] == 255:
+        @run_to_nodes
+        def nodes(data):
+            nonlocal n
+            data.draw_bytes(n, n)
+            data.draw_bytes(255 - n, 255 - n)
+            if n == 255:
                 data.mark_interesting()
             else:
-                n[0] += 1
+                n += 1
 
 
 def test_no_read_no_shrink():
-    count = [0]
+    count = 0
 
-    @run_to_buffer
-    def x(data):
-        count[0] += 1
+    @run_to_nodes
+    def nodes(data):
+        nonlocal count
+        count += 1
         data.mark_interesting()
 
-    assert x == b""
-    assert count == [1]
+    assert nodes == ()
+    assert count == 1
 
 
 def test_one_dead_branch():
     with deterministic_PRNG():
         seen = set()
 
-        @run_to_buffer
-        def x(data):
+        @run_to_nodes
+        def nodes(data):
             i = data.draw_bytes(1, 1)[0]
             if i > 0:
                 data.mark_invalid()
@@ -394,15 +401,15 @@ def test_does_not_save_on_interrupt():
     assert not db.data
 
 
-def test_returns_written():
+def test_returns_forced():
     value = b"\0\1\2\3"
 
-    @run_to_buffer
-    def written(data):
+    @run_to_nodes
+    def nodes(data):
         data.draw_bytes(len(value), len(value), forced=value)
         data.mark_interesting()
 
-    assert value == written
+    assert tuple(n.value for n in nodes) == (value,)
 
 
 def fails_health_check(label, **kwargs):
@@ -482,6 +489,7 @@ def test_can_shrink_variable_string_draws():
 def test_variable_size_string_increasing():
     # coverage test for min_size increasing during shrinking (because the test
     # function inverts n).
+    # ...except this currently overruns instead and misses that check.
     @st.composite
     def strategy(draw):
         n = 10 - draw(st.integers(0, 10))
@@ -506,11 +514,11 @@ class Foo:
 
 
 def test_debug_data(capsys):
-    buf = [0, 1, 2]
+    nodes = ir(0, 1, 2)
 
     def f(data):
-        for x in bytes(buf):
-            if data.draw(st.integers(0, 100)) != x:
+        for node in nodes:
+            if data.draw(st.integers(0, 100)) != node.value:
                 data.mark_invalid()
             data.start_example(1)
             data.stop_example()
@@ -525,11 +533,11 @@ def test_debug_data(capsys):
             verbosity=Verbosity.debug,
         ),
     )
-    runner.cached_test_function(buf)
+    runner.cached_test_function_ir(nodes)
     runner.run()
 
     out, _ = capsys.readouterr()
-    assert re.match("\\d+ bytes \\[.*\\] -> ", out)
+    assert re.match(r"\d+ choices \(.*\) -> ", out)
     assert "INTERESTING" in out
 
 
@@ -561,12 +569,12 @@ def test_uniqueness_is_preserved_when_writing_at_beginning():
 
 
 @pytest.mark.parametrize("skip_target", [False, True])
-@pytest.mark.parametrize("initial_attempt", [127, 128])
+@pytest.mark.parametrize("initial_attempt", [ir(127, 128)])
 def test_clears_out_its_database_on_shrinking(
     initial_attempt, skip_target, monkeypatch
 ):
     def generate_new_examples(self):
-        self.cached_test_function(initial_attempt)
+        self.cached_test_function_ir(initial_attempt)
 
     monkeypatch.setattr(
         ConjectureRunner, "generate_new_examples", generate_new_examples
@@ -597,26 +605,26 @@ def test_clears_out_its_database_on_shrinking(
 
 
 def test_detects_too_small_block_starts():
-    call_count = [0]
+    call_count = 0
 
     def f(data):
-        assert call_count[0] == 0
-        call_count[0] += 1
+        nonlocal call_count
+        call_count += 1
         data.draw_bytes(8, 8)
         data.mark_interesting()
 
     runner = ConjectureRunner(f, settings=settings(database=None))
     r = runner.cached_test_function(bytes(8))
-    assert r.status == Status.INTERESTING
-    assert call_count[0] == 1
+    assert r.status is Status.INTERESTING
+    assert call_count == 1
     r2 = runner.cached_test_function(bytes([255] * 7))
-    assert r2.status == Status.OVERRUN
-    assert call_count[0] == 1
+    assert r2.status is Status.OVERRUN
+    assert call_count == 1
 
 
 def test_shrinks_both_interesting_examples(monkeypatch):
     def generate_new_examples(self):
-        self.cached_test_function(bytes([1]))
+        self.cached_test_function_ir(ir(1))
 
     monkeypatch.setattr(
         ConjectureRunner, "generate_new_examples", generate_new_examples
@@ -628,8 +636,8 @@ def test_shrinks_both_interesting_examples(monkeypatch):
 
     runner = ConjectureRunner(f, database_key=b"key")
     runner.run()
-    assert runner.interesting_examples[0].buffer == bytes([0])
-    assert runner.interesting_examples[1].buffer == bytes([1])
+    assert runner.interesting_examples[0].choices == (0,)
+    assert runner.interesting_examples[1].choices == (1,)
 
 
 def test_discarding(monkeypatch):
@@ -637,11 +645,11 @@ def test_discarding(monkeypatch):
     monkeypatch.setattr(
         ConjectureRunner,
         "generate_new_examples",
-        lambda runner: runner.cached_test_function(bytes([0, 1] * 10)),
+        lambda runner: runner.cached_test_function_ir(ir(False, True) * 10),
     )
 
-    @run_to_buffer
-    def x(data):
+    @run_to_nodes
+    def nodes(data):
         count = 0
         while count < 10:
             data.start_example(SOME_LABEL)
@@ -651,12 +659,12 @@ def test_discarding(monkeypatch):
             data.stop_example(discard=not b)
         data.mark_interesting()
 
-    assert x == bytes(bytes([1]) * 10)
+    assert tuple(n.value for n in nodes) == (True,) * 10
 
 
 def test_can_remove_discarded_data():
-    @shrinking_from(bytes([0] * 10 + [11]))
-    def shrinker(data):
+    @shrinking_from(ir(0) * 10 + ir(11))
+    def shrinker(data: ConjectureData):
         while True:
             data.start_example(SOME_LABEL)
             b = data.draw_integer(0, 2**8 - 1)
@@ -666,12 +674,12 @@ def test_can_remove_discarded_data():
         data.mark_interesting()
 
     shrinker.remove_discarded()
-    assert list(shrinker.buffer) == [11]
+    assert shrinker.choices == (11,)
 
 
 def test_discarding_iterates_to_fixed_point():
-    @shrinking_from(bytes(list(range(100, -1, -1))))
-    def shrinker(data):
+    @shrinking_from(ir(*list(range(100, -1, -1))))
+    def shrinker(data: ConjectureData):
         data.start_example(0)
         data.draw_integer(0, 2**8 - 1)
         data.stop_example(discard=True)
@@ -680,12 +688,12 @@ def test_discarding_iterates_to_fixed_point():
         data.mark_interesting()
 
     shrinker.remove_discarded()
-    assert list(shrinker.buffer) == [1, 0]
+    assert shrinker.choices == (1, 0)
 
 
 def test_discarding_is_not_fooled_by_empty_discards():
-    @shrinking_from(bytes([1, 1]))
-    def shrinker(data):
+    @shrinking_from(ir(1, 1))
+    def shrinker(data: ConjectureData):
         data.draw_integer(0, 2**1 - 1)
         data.start_example(0)
         data.stop_example(discard=True)
@@ -696,9 +704,9 @@ def test_discarding_is_not_fooled_by_empty_discards():
     assert shrinker.shrink_target.has_discards
 
 
-def test_discarding_can_fail(monkeypatch):
-    @shrinking_from(bytes([1]))
-    def shrinker(data):
+def test_discarding_can_fail():
+    @shrinking_from(ir(1))
+    def shrinker(data: ConjectureData):
         data.start_example(0)
         data.draw_boolean()
         data.stop_example(discard=True)
@@ -712,16 +720,16 @@ def test_shrinking_from_mostly_zero(monkeypatch):
     monkeypatch.setattr(
         ConjectureRunner,
         "generate_new_examples",
-        lambda self: self.cached_test_function(bytes(5) + bytes([2])),
+        lambda self: self.cached_test_function_ir(ir(0) * 5 + ir(2)),
     )
 
-    @run_to_buffer
-    def x(data):
+    @run_to_nodes
+    def nodes(data):
         s = [data.draw_integer(0, 2**8 - 1) for _ in range(6)]
         if any(s):
             data.mark_interesting()
 
-    assert x == bytes(5) + bytes([1])
+    assert tuple(n.value for n in nodes) == (0,) * 5 + (1,)
 
 
 def test_handles_nesting_of_discard_correctly(monkeypatch):
@@ -729,11 +737,11 @@ def test_handles_nesting_of_discard_correctly(monkeypatch):
     monkeypatch.setattr(
         ConjectureRunner,
         "generate_new_examples",
-        lambda runner: runner.cached_test_function(bytes([0, 0, 1, 1])),
+        lambda runner: runner.cached_test_function_ir(ir(False, False, True, True)),
     )
 
-    @run_to_buffer
-    def x(data):
+    @run_to_nodes
+    def nodes(data):
         while True:
             data.start_example(SOME_LABEL)
             succeeded = data.draw_boolean()
@@ -744,7 +752,7 @@ def test_handles_nesting_of_discard_correctly(monkeypatch):
             if succeeded:
                 data.mark_interesting()
 
-    assert x == bytes([1, 1])
+    assert tuple(n.value for n in nodes) == (True, True)
 
 
 def test_database_clears_secondary_key():
@@ -768,7 +776,7 @@ def test_database_clears_secondary_key():
     for i in range(10):
         database.save(runner.secondary_key, bytes([i]))
 
-    runner.cached_test_function([10])
+    runner.cached_test_function_ir(ir(10))
     assert runner.interesting_examples
 
     assert len(set(database.fetch(key))) == 1
@@ -797,13 +805,11 @@ def test_database_uses_values_from_secondary_key():
         ),
         database_key=key,
     )
-
     for i in range(10):
         database.save(runner.secondary_key, bytes([i]))
 
-    runner.cached_test_function([10])
+    runner.cached_test_function_ir(ir(10))
     assert runner.interesting_examples
-
     assert len(set(database.fetch(key))) == 1
     assert len(set(database.fetch(runner.secondary_key))) == 10
 
@@ -815,7 +821,6 @@ def test_database_uses_values_from_secondary_key():
     )
 
     (v,) = runner.interesting_examples.values()
-
     assert list(v.buffer) == [5]
 
 
@@ -838,11 +843,12 @@ def test_exit_because_max_iterations():
 
 
 def test_exit_because_shrink_phase_timeout(monkeypatch):
-    val = [0]
+    val = 0
 
     def fast_time():
-        val[0] += 1000
-        return val[0]
+        nonlocal val
+        val += 1000
+        return val
 
     def f(data):
         if data.draw_integer(0, 2**64 - 1) > 2**33:
@@ -856,8 +862,8 @@ def test_exit_because_shrink_phase_timeout(monkeypatch):
 
 
 def test_dependent_block_pairs_can_lower_to_zero():
-    @shrinking_from([1, 0, 1])
-    def shrinker(data):
+    @shrinking_from(ir(True, 1))
+    def shrinker(data: ConjectureData):
         if data.draw_boolean():
             n = data.draw_integer(0, 2**16 - 1)
         else:
@@ -867,12 +873,12 @@ def test_dependent_block_pairs_can_lower_to_zero():
             data.mark_interesting()
 
     shrinker.fixate_shrink_passes(["minimize_individual_nodes"])
-    assert list(shrinker.shrink_target.buffer) == [0, 1]
+    assert shrinker.choices == (False, 1)
 
 
 def test_handle_size_too_large_during_dependent_lowering():
-    @shrinking_from([1, 255, 0])
-    def shrinker(data):
+    @shrinking_from(ir(True, 255, 0))
+    def shrinker(data: ConjectureData):
         if data.draw_boolean():
             data.draw_integer(0, 2**16 - 1)
             data.mark_interesting()
@@ -883,10 +889,8 @@ def test_handle_size_too_large_during_dependent_lowering():
 
 
 def test_block_may_grow_during_lexical_shrinking():
-    initial = bytes([2, 1, 1])
-
-    @shrinking_from(initial)
-    def shrinker(data):
+    @shrinking_from(ir(2, 1, 1))
+    def shrinker(data: ConjectureData):
         n = data.draw_integer(0, 2**8 - 1)
         if n == 2:
             data.draw_integer(0, 2**8 - 1)
@@ -896,12 +900,12 @@ def test_block_may_grow_during_lexical_shrinking():
         data.mark_interesting()
 
     shrinker.fixate_shrink_passes(["minimize_individual_nodes"])
-    assert list(shrinker.shrink_target.buffer) == [0, 0, 0]
+    assert shrinker.choices == (0, 0)
 
 
 def test_lower_common_node_offset_does_nothing_when_changed_blocks_are_zero():
-    @shrinking_from([1, 0, 1, 0])
-    def shrinker(data):
+    @shrinking_from(ir(True, False, True, False))
+    def shrinker(data: ConjectureData):
         data.draw_boolean()
         data.draw_boolean()
         data.draw_boolean()
@@ -911,12 +915,12 @@ def test_lower_common_node_offset_does_nothing_when_changed_blocks_are_zero():
     shrinker.mark_changed(1)
     shrinker.mark_changed(3)
     shrinker.lower_common_node_offset()
-    assert list(shrinker.shrink_target.buffer) == [1, 0, 1, 0]
+    assert shrinker.choices == (True, False, True, False)
 
 
 def test_lower_common_node_offset_ignores_zeros():
-    @shrinking_from([2, 2, 0])
-    def shrinker(data):
+    @shrinking_from(ir(2, 2, 0))
+    def shrinker(data: ConjectureData):
         n = data.draw_integer(0, 2**8 - 1)
         data.draw_integer(0, 2**8 - 1)
         data.draw_integer(0, 2**8 - 1)
@@ -926,32 +930,34 @@ def test_lower_common_node_offset_ignores_zeros():
     for i in range(3):
         shrinker.mark_changed(i)
     shrinker.lower_common_node_offset()
-    assert list(shrinker.shrink_target.buffer) == [1, 1, 0]
+    assert shrinker.choices == (1, 1, 0)
 
 
 def test_cached_test_function_returns_right_value():
-    count = [0]
+    count = 0
 
     def tf(data):
-        count[0] += 1
+        nonlocal count
+        count += 1
         data.draw_integer(0, 3)
         data.mark_interesting()
 
     with deterministic_PRNG():
         runner = ConjectureRunner(tf, settings=TEST_SETTINGS)
         for _ in range(2):
-            for b in (b"\0", b"\1"):
-                d = runner.cached_test_function(b)
+            for nodes in (ir(0), ir(1)):
+                d = runner.cached_test_function_ir(nodes)
                 assert d.status == Status.INTERESTING
-                assert d.buffer == b
-        assert count[0] == 2
+                assert d.choices == (nodes[0].value,)
+        assert count == 2
 
 
 def test_cached_test_function_does_not_reinvoke_on_prefix():
-    call_count = [0]
+    call_count = 0
 
     def test_function(data):
-        call_count[0] += 1
+        nonlocal call_count
+        call_count += 1
         data.draw_integer(0, 2**8 - 1)
         data.draw_bytes(1, 1, forced=bytes([7]))
         data.draw_integer(0, 2**8 - 1)
@@ -959,32 +965,33 @@ def test_cached_test_function_does_not_reinvoke_on_prefix():
     with deterministic_PRNG():
         runner = ConjectureRunner(test_function, settings=TEST_SETTINGS)
 
-        data = runner.cached_test_function(bytes(3))
+        data = runner.cached_test_function_ir(ir(0, b"\0", 0))
         assert data.status == Status.VALID
         for n in [2, 1, 0]:
-            prefix_data = runner.cached_test_function(bytes(n))
-            assert prefix_data is Overrun
-        assert call_count[0] == 1
+            d = runner.cached_test_function_ir(data.ir_nodes[:n])
+            assert d is Overrun
+        assert call_count == 1
 
 
 def test_will_evict_entries_from_the_cache(monkeypatch):
     monkeypatch.setattr(engine_module, "CACHE_SIZE", 5)
-    count = [0]
+    count = 0
 
     def tf(data):
-        data.draw_bytes(1, 1)
-        count[0] += 1
+        nonlocal count
+        data.draw_integer(0, 2**8 - 1)
+        count += 1
 
     runner = ConjectureRunner(tf, settings=TEST_SETTINGS)
 
     for _ in range(3):
         for n in range(10):
-            runner.cached_test_function([n])
+            runner.cached_test_function_ir(ir(n))
 
     # Because we exceeded the cache size, our previous
     # calls will have been evicted, so each call to
     # cached_test_function will have to reexecute.
-    assert count[0] == 30
+    assert count == 30
 
 
 def test_branch_ending_in_write():
@@ -998,19 +1005,18 @@ def test_branch_ending_in_write():
         if count > 1:
             data.draw_boolean(forced=False)
 
-        b = bytes(data.buffer)
-        assert b not in seen
-        seen.add(b)
+        assert data.ir_nodes not in seen
+        seen.add(data.ir_nodes)
 
     with deterministic_PRNG():
         runner = ConjectureRunner(tf, settings=TEST_SETTINGS)
 
         for _ in range(100):
             prefix = runner.generate_novel_prefix()
-            attempt = prefix + bytes(2)
-            data = runner.cached_test_function(attempt)
-            assert data.status == Status.VALID
-            assert attempt.startswith(data.buffer)
+            attempt = prefix + ir(False, False)
+            data = runner.cached_test_function_ir(attempt)
+            assert data.status is Status.VALID
+            assert startswith(tuple(n.value for n in attempt), data.choices)
 
 
 def test_exhaust_space():
@@ -1091,7 +1097,7 @@ def test_does_not_shrink_multiple_bugs_when_told_not_to():
         runner = ConjectureRunner(
             test, settings=settings(TEST_SETTINGS, report_multiple_bugs=False)
         )
-        runner.cached_test_function([255, 255])
+        runner.cached_test_function_ir(ir(255, 255))
         runner.shrink_interesting_examples()
 
         results = {d.buffer for d in runner.interesting_examples.values()}
@@ -1285,15 +1291,15 @@ def test_replaces_all_dominated():
         database_key=b"stuff",
     )
 
-    d1 = runner.cached_test_function([0, 1]).as_result()
-    d2 = runner.cached_test_function([1, 0]).as_result()
+    d1 = runner.cached_test_function_ir(ir(0, 1)).as_result()
+    d2 = runner.cached_test_function_ir(ir(1, 0)).as_result()
 
     assert len(runner.pareto_front) == 2
 
     assert runner.pareto_front[0] == d1
     assert runner.pareto_front[1] == d2
 
-    d3 = runner.cached_test_function([0, 0]).as_result()
+    d3 = runner.cached_test_function_ir(ir(0, 0)).as_result()
     assert len(runner.pareto_front) == 1
 
     assert runner.pareto_front[0] == d3
@@ -1308,18 +1314,13 @@ def test_does_not_duplicate_elements():
         settings=settings(TEST_SETTINGS, database=InMemoryExampleDatabase()),
         database_key=b"stuff",
     )
-
-    d1 = runner.cached_test_function([1]).as_result()
+    d1 = runner.cached_test_function_ir(ir(1)).as_result()
 
     assert len(runner.pareto_front) == 1
-
     # This can happen in practice if we e.g. reexecute a test because it has
     # expired from the cache. It's easier just to test it directly though
     # rather than simulate the failure mode.
-    is_pareto = runner.pareto_front.add(d1)
-
-    assert is_pareto
-
+    assert runner.pareto_front.add(d1)
     assert len(runner.pareto_front) == 1
 
 
@@ -1334,8 +1335,8 @@ def test_includes_right_hand_side_targets_in_dominance():
         database_key=b"stuff",
     )
 
-    d1 = runner.cached_test_function([0]).as_result()
-    d2 = runner.cached_test_function([1]).as_result()
+    d1 = runner.cached_test_function_ir(ir(0)).as_result()
+    d2 = runner.cached_test_function_ir(ir(1)).as_result()
 
     assert dominance(d1, d2) == DominanceRelation.NO_DOMINANCE
 
@@ -1351,8 +1352,8 @@ def test_smaller_interesting_dominates_larger_valid():
         database_key=b"stuff",
     )
 
-    d1 = runner.cached_test_function([0]).as_result()
-    d2 = runner.cached_test_function([1]).as_result()
+    d1 = runner.cached_test_function_ir(ir(0)).as_result()
+    d2 = runner.cached_test_function_ir(ir(1)).as_result()
     assert dominance(d1, d2) == DominanceRelation.LEFT_DOMINATES
 
 
@@ -1378,9 +1379,7 @@ def test_runs_optimisation_even_if_not_generating():
         runner = ConjectureRunner(
             test, settings=settings(TEST_SETTINGS, phases=[Phase.target])
         )
-
-        runner.cached_test_function(bytes(2))
-
+        runner.cached_test_function_ir(ir(0))
         runner.run()
 
         assert runner.best_observed_targets["n"] == (2**16) - 1
@@ -1422,50 +1421,53 @@ def test_does_not_run_optimisation_when_max_examples_is_small():
 
 def test_does_not_cache_extended_prefix():
     def test(data):
-        data.draw_bytes(8, 8)
+        data.draw_integer()
+        data.draw_integer()
 
     with deterministic_PRNG():
         runner = ConjectureRunner(test, settings=TEST_SETTINGS)
 
-        d1 = runner.cached_test_function(b"", extend=8)
-        d2 = runner.cached_test_function(b"", extend=8)
-        assert d1.status == d2.status == Status.VALID
-
-        assert d1.buffer != d2.buffer
+        d1 = runner.cached_test_function_ir(ir(0), extend=10)
+        assert runner.call_count == 1
+        d2 = runner.cached_test_function_ir(ir(0), extend=10)
+        assert runner.call_count == 2
+        assert d1.status is d2.status is Status.VALID
 
 
 def test_does_cache_if_extend_is_not_used():
-    calls = [0]
+    calls = 0
 
     def test(data):
-        calls[0] += 1
+        nonlocal calls
+        calls += 1
         data.draw_bytes(1, 1)
 
     with deterministic_PRNG():
         runner = ConjectureRunner(test, settings=TEST_SETTINGS)
 
-        d1 = runner.cached_test_function(b"\0", extend=8)
-        d2 = runner.cached_test_function(b"\0", extend=8)
+        d1 = runner.cached_test_function_ir(ir(b"\0"), extend=8)
+        d2 = runner.cached_test_function_ir(ir(b"\0"), extend=8)
         assert d1.status == d2.status == Status.VALID
         assert d1.buffer == d2.buffer
-        assert calls[0] == 1
+        assert calls == 1
 
 
 def test_does_result_for_reuse():
-    calls = [0]
+    calls = 0
 
     def test(data):
-        calls[0] += 1
+        nonlocal calls
+        calls += 1
         data.draw_bytes(1, 1)
 
     with deterministic_PRNG():
         runner = ConjectureRunner(test, settings=TEST_SETTINGS)
 
-        d1 = runner.cached_test_function(b"\0", extend=8)
-        d2 = runner.cached_test_function(d1.buffer)
+        d1 = runner.cached_test_function_ir(ir(b"\0"), extend=8)
+        d2 = runner.cached_test_function_ir(d1.ir_nodes)
         assert d1.status == d2.status == Status.VALID
-        assert d1.buffer == d2.buffer
-        assert calls[0] == 1
+        assert d1.ir_nodes == d2.ir_nodes
+        assert calls == 1
 
 
 def test_does_not_cache_overrun_if_extending():
@@ -1475,8 +1477,8 @@ def test_does_not_cache_overrun_if_extending():
     with deterministic_PRNG():
         runner = ConjectureRunner(test, settings=TEST_SETTINGS)
 
-        d1 = runner.cached_test_function(b"", extend=4)
-        d2 = runner.cached_test_function(b"", extend=8)
+        d1 = runner.cached_test_function_ir(ir(b""), extend=4)
+        d2 = runner.cached_test_function_ir(ir(b""), extend=8)
         assert d1.status == Status.OVERRUN
         assert d2.status == Status.VALID
 
@@ -1489,10 +1491,11 @@ def test_does_cache_overrun_if_not_extending():
     with deterministic_PRNG():
         runner = ConjectureRunner(test, settings=TEST_SETTINGS)
 
-        d1 = runner.cached_test_function(bytes(8), extend=0)
-        d2 = runner.cached_test_function(bytes(8), extend=8)
-        assert d1.status == Status.OVERRUN
-        assert d2.status == Status.VALID
+        data = runner.cached_test_function_ir(ir(bytes(8)), extend=0)
+        assert data.status is Status.OVERRUN
+
+        data = runner.cached_test_function_ir(ir(bytes(8)), extend=10)
+        assert data.status is Status.VALID
 
 
 def test_does_not_cache_extended_prefix_if_overrun():
@@ -1502,23 +1505,35 @@ def test_does_not_cache_extended_prefix_if_overrun():
     with deterministic_PRNG():
         runner = ConjectureRunner(test, settings=TEST_SETTINGS)
 
-        d1 = runner.cached_test_function(b"", extend=4)
-        d2 = runner.cached_test_function(b"", extend=8)
+        d1 = runner.cached_test_function_ir(ir(b""), extend=4)
+        d2 = runner.cached_test_function_ir(ir(b""), extend=8)
         assert d1.status == Status.OVERRUN
         assert d2.status == Status.VALID
 
 
+def test_draw_bits_partly_from_prefix_and_partly_random():
+    # a draw_bits call which straddles the end of our prefix has a slightly
+    # different code branch.
+    def test(data):
+        # float consumes draw_bits(64)
+        data.draw_float()
+
+    with deterministic_PRNG():
+        runner = ConjectureRunner(test, settings=TEST_SETTINGS)
+        d = runner.cached_test_function(bytes(10), extend=100)
+        assert d.status == Status.VALID
+
+
 def test_can_be_set_to_ignore_limits():
     def test(data):
-        data.draw_bytes(1, 1)
+        data.draw_integer(0, 2**8 - 1)
 
     with deterministic_PRNG():
         runner = ConjectureRunner(
             test, settings=settings(TEST_SETTINGS, max_examples=1), ignore_limits=True
         )
-
         for c in range(256):
-            runner.cached_test_function([c])
+            runner.cached_test_function_ir(ir(c))
 
         assert runner.tree.is_exhausted
 
@@ -1578,6 +1593,34 @@ def test_cache_ignores_was_forced(forced_first, node):
     assert runner.call_count == 1
 
 
+@given(ir_nodes(was_forced=False))
+def test_overruns_with_extend_are_not_cached(node):
+    assume(compute_max_children(node.ir_type, node.kwargs) > 100)
+
+    def test(cd):
+        _draw(cd, node)
+        _draw(cd, node)
+
+    runner = ConjectureRunner(test)
+    assert runner.call_count == 0
+
+    data = runner.cached_test_function_ir([node])
+    assert runner.call_count == 1
+    assert data.status is Status.OVERRUN
+
+    # cache hit
+    data = runner.cached_test_function_ir([node])
+    assert runner.call_count == 1
+    assert data.status is Status.OVERRUN
+
+    # cache miss
+    data = runner.cached_test_function_ir(
+        [node], extend=BUFFER_SIZE_IR - ir_size_nodes([node])
+    )
+    assert runner.call_count == 2
+    assert data.status is Status.VALID
+
+
 def test_simulate_to_evicted_data(monkeypatch):
     # test that we do not rely on the false invariant that correctly simulating
     # a data to a result means we have that result in the cache, due to e.g.
@@ -1631,3 +1674,72 @@ def test_mildly_complicated_strategies(strategy, condition):
     # covered by shrinking any mildly compliated strategy and aren't worth
     # testing explicitly for. This covers those.
     minimal(strategy, condition)
+
+
+def test_does_not_shrink_if_replaying_from_database():
+    db = InMemoryExampleDatabase()
+    key = b"foo"
+
+    def f(data):
+        if data.draw_integer(0, 255) == 123:
+            data.mark_interesting()
+
+    runner = ConjectureRunner(f, settings=settings(database=db), database_key=key)
+    b = bytes([123])
+    runner.save_buffer(b)
+    runner.shrink_interesting_examples = None
+    runner.run()
+    (last_data,) = runner.interesting_examples.values()
+    assert last_data.buffer == b
+
+
+def test_does_shrink_if_replaying_inexact_from_database():
+    db = InMemoryExampleDatabase()
+    key = b"foo"
+
+    def f(data):
+        data.draw_integer(0, 255)
+        data.mark_interesting()
+
+    runner = ConjectureRunner(f, settings=settings(database=db), database_key=key)
+    b = bytes([123, 2])
+    runner.save_buffer(b)
+    runner.run()
+    (last_data,) = runner.interesting_examples.values()
+    assert last_data.buffer == bytes([0])
+
+
+def test_stops_if_hits_interesting_early_and_only_want_one_bug():
+    db = InMemoryExampleDatabase()
+    key = b"foo"
+
+    def f(data):
+        data.draw_integer(0, 255)
+        data.mark_interesting()
+
+    runner = ConjectureRunner(
+        f, settings=settings(database=db, report_multiple_bugs=False), database_key=key
+    )
+    for i in range(256):
+        runner.save_buffer(bytes([i]))
+    runner.run()
+    assert runner.call_count == 1
+
+
+def test_skips_secondary_if_interesting_is_found():
+    db = InMemoryExampleDatabase()
+    key = b"foo"
+
+    def f(data):
+        data.draw_integer(0, 255)
+        data.mark_interesting()
+
+    runner = ConjectureRunner(
+        f,
+        settings=settings(max_examples=1000, database=db, report_multiple_bugs=True),
+        database_key=key,
+    )
+    for i in range(256):
+        db.save(runner.database_key if i < 10 else runner.secondary_key, bytes([i]))
+    runner.reuse_existing_examples()
+    assert runner.call_count == 10
