@@ -19,7 +19,7 @@ from hypothesis.internal.conjecture.data import ConjectureData
 from hypothesis.internal.conjecture.engine import ConjectureRunner
 from hypothesis.internal.floats import float_to_int
 
-from tests.conjecture.common import run_to_buffer
+from tests.conjecture.common import ir
 
 EXPONENTS = list(range(flt.MAX_EXPONENT + 1))
 assert len(EXPONENTS) == 2**11
@@ -80,7 +80,7 @@ def test_floats_round_trip(f):
     assert float_to_int(f) == float_to_int(g)
 
 
-@settings(suppress_health_check=[HealthCheck.too_slow])
+@settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
 @example(1, 0.5)
 @given(st.integers(1, 2**53), st.floats(0, 1).filter(lambda x: x not in (0, 1)))
 def test_floats_order_worse_than_their_integral_part(n, g):
@@ -123,29 +123,28 @@ def test_reverse_bits_table_has_right_elements():
     assert sorted(flt.REVERSE_BITS_TABLE) == list(range(256))
 
 
-def float_runner(start, condition):
-    @run_to_buffer
-    def buf(data):
-        data.draw_float(forced=start)
-        data.mark_interesting()
+def float_runner(start, condition, *, kwargs=None):
+    kwargs = {} if kwargs is None else kwargs
 
     def test_function(data):
-        f = data.draw_float()
+        f = data.draw_float(**kwargs)
         if condition(f):
             data.mark_interesting()
 
     runner = ConjectureRunner(test_function)
-    runner.cached_test_function(buf)
+    runner.cached_test_function_ir(ir((float(start), kwargs)))
     assert runner.interesting_examples
     return runner
 
 
-def minimal_from(start, condition):
-    runner = float_runner(start, condition)
+def minimal_from(start, condition, *, kwargs=None):
+    kwargs = {} if kwargs is None else kwargs
+
+    runner = float_runner(start, condition, kwargs=kwargs)
     runner.shrink_interesting_examples()
     (v,) = runner.interesting_examples.values()
-    data = ConjectureData.for_buffer(v.buffer)
-    result = data.draw_float()
+    data = ConjectureData.for_ir_tree(v.ir_nodes)
+    result = data.draw_float(**kwargs)
     assert condition(result)
     return result
 
@@ -211,3 +210,10 @@ def test_converts_floats_to_integer_form(f):
     runner.shrink_interesting_examples()
     (v,) = runner.interesting_examples.values()
     assert v.buffer[:-1] < buf
+
+
+def test_reject_out_of_bounds_floats_while_shrinking():
+    # coverage test for rejecting out of bounds floats while shrinking
+    kwargs = {"min_value": 103.0}
+    g = minimal_from(103.1, lambda x: x >= 100, kwargs=kwargs)
+    assert g == 103.0
