@@ -28,14 +28,11 @@ def fails_with_output(expected, error=AssertionError, **kw):
     return _inner
 
 
-# this should have a marked as freely varying, but false negatives in our
-# inquisitor code skip over it sometimes, depending on the seen_passed_buffers.
-# yet another thing that should be improved by moving to the ir.
 @fails_with_output(
     """
 Falsifying example: test_inquisitor_comments_basic_fail_if_either(
     # The test always failed when commented parts were varied together.
-    a=False,
+    a=False,  # or any other generated value
     b=True,
     c=[],  # or any other generated value
     d=True,
@@ -77,6 +74,27 @@ def test_inquisitor_no_together_comment_if_single_argument(a, b):
     assert a
 
 
+@st.composite
+def ints_with_forced_draw(draw):
+    data = draw(st.data())
+    n = draw(st.integers())
+    data.conjecture_data.draw_boolean(forced=True)
+    return n
+
+
+@fails_with_output(
+    """
+Falsifying example: test_inquisitor_doesnt_break_on_varying_forced_nodes(
+    n1=100,
+    n2=0,  # or any other generated value
+)
+"""
+)
+@given(st.integers(), ints_with_forced_draw())
+def test_inquisitor_doesnt_break_on_varying_forced_nodes(n1, n2):
+    assert n1 < 100
+
+
 @fails_with(ZeroDivisionError)
 @settings(database=None)
 @given(start_date=st.datetimes(), data=st.data())
