@@ -20,7 +20,9 @@ from hypothesis.internal.conjecture.data import (
     COLLECTION_DEFAULT_MAX_SIZE,
     ConjectureData,
     IRNode,
+    NodeTemplate,
     Status,
+    ir_size_nodes,
     ir_value_equal,
     ir_value_permitted,
 )
@@ -29,6 +31,7 @@ from hypothesis.internal.conjecture.datatree import (
     all_children,
     compute_max_children,
 )
+from hypothesis.internal.conjecture.engine import BUFFER_SIZE_IR, truncate_nodes_to_size
 from hypothesis.internal.floats import SMALLEST_SUBNORMAL, next_down, next_up
 from hypothesis.internal.intervalsets import IntervalSet
 
@@ -37,6 +40,7 @@ from tests.conjecture.common import (
     draw_value,
     fresh_data,
     integer_kwargs,
+    ir,
     ir_nodes,
     ir_types_and_kwargs,
 )
@@ -333,7 +337,7 @@ def test_ir_nodes(random):
     data.draw_integer(0, 100, forced=50)
 
     data.freeze()
-    expected_tree_nodes = [
+    expected_tree_nodes = (
         IRNode(
             ir_type="float",
             value=5.0,
@@ -378,8 +382,8 @@ def test_ir_nodes(random):
             },
             was_forced=True,
         ),
-    ]
-    assert data.examples.ir_tree_nodes == expected_tree_nodes
+    )
+    assert data.ir_nodes == expected_tree_nodes
 
 
 @given(ir_nodes())
@@ -401,6 +405,12 @@ def test_ir_node_equality(node):
     assert node != 42
 
 
+@given(ir_nodes(was_forced=True))
+def test_cannot_modify_forced_nodes(node):
+    with pytest.raises(AssertionError):
+        node.copy(with_value=42)
+
+
 def test_data_with_empty_ir_tree_is_overrun():
     data = ConjectureData.for_ir_tree([])
     with pytest.raises(StopTest):
@@ -418,7 +428,7 @@ def test_data_with_changed_forced_value(node):
     # This is actually fine; we'll just ignore the forced node (v1) and return
     # what the draw expects (v2).
 
-    data = ConjectureData.for_ir_tree([node])
+    data = ConjectureData.for_ir_tree([node], max_length=BUFFER_SIZE_IR)
 
     draw_func = getattr(data, f"draw_{node.ir_type}")
     kwargs = deepcopy(node.kwargs)
@@ -669,6 +679,18 @@ def test_forced_nodes_are_trivial(node):
             kwargs={"p": 0.5},
             was_forced=False,
         ),
+        IRNode(
+            ir_type="boolean",
+            value=True,
+            kwargs={"p": 1.0},
+            was_forced=False,
+        ),
+        IRNode(
+            ir_type="boolean",
+            value=False,
+            kwargs={"p": 0.0},
+            was_forced=False,
+        ),
         IRNode(
             ir_type="string",
             value="",
@@ -767,6 +789,41 @@ def test_forced_nodes_are_trivial(node):
             },
             was_forced=False,
         ),
+        IRNode(
+            ir_type="integer",
+            value=1,
+            kwargs={
+                "min_value": -10,
+                "max_value": None,
+                "weights": None,
+                "shrink_towards": 1,
+            },
+            was_forced=False,
+        ),
+        IRNode(
+            ir_type="integer",
+            value=1,
+            kwargs={
+                "min_value": None,
+                "max_value": 10,
+                "weights": None,
+                "shrink_towards": 1,
+            },
+            was_forced=False,
+        ),
+        # we don't consider shrink_towards for unbounded integers.
+        # the trivial value should probably be 1 here, not 0.
+        IRNode(
+            ir_type="integer",
+            value=0,
+            kwargs={
+                "min_value": None,
+                "max_value": None,
+                "weights": None,
+                "shrink_towards": 1,
+            },
+            was_forced=False,
+        ),
     ],
 )
 def test_trivial_nodes(node):
@@ -823,6 +880,12 @@ def test_trivial_nodes(node):
             kwargs={"p": 0.5},
             was_forced=False,
         ),
+        IRNode(
+            ir_type="boolean",
+            value=True,
+            kwargs={"p": 0.99},
+            was_forced=False,
+        ),
         IRNode(
             ir_type="string",
             value="d",
@@ -963,3 +1026,49 @@ def test_conservative_nontrivial_nodes(node):
 @given(ir_nodes())
 def test_ir_node_is_hashable(ir_node):
     hash(ir_node)
+
+
+@given(st.lists(ir_nodes()))
+def test_ir_size_positive(nodes):
+    assert ir_size_nodes(nodes) >= 0
+
+
+@given(st.integers(min_value=1))
+def test_node_template_size(n):
+    node = NodeTemplate(type="simplest", size=n)
+    assert ir_size_nodes([node]) == n
+
+
+@given(st.lists(ir_nodes()), st.integers(min_value=0))
+def test_truncate_nodes(nodes, size):
+    assert len(truncate_nodes_to_size(nodes, size)) <= len(nodes)
+
+
+def test_node_template_to_overrun():
+    data = ConjectureData.for_ir_tree(ir(1) + (NodeTemplate("simplest", size=10),))
+    data.draw_integer()
+    with pytest.raises(StopTest):
+        for _ in range(10):
+            data.draw_integer()
+
+    assert data.status is Status.OVERRUN
+
+
+def test_node_template_single_node_overruns():
+    # test for when drawing a single node takes more than BUFFER_SIZE, while in
+    # the NodeTemplate case
+    data = ConjectureData.for_ir_tree((NodeTemplate("simplest", size=BUFFER_SIZE_IR),))
+    with pytest.raises(StopTest):
+        data.draw_bytes(10_000, 10_000)
+
+    assert data.status is Status.OVERRUN
+
+
+@given(ir_nodes())
+def test_node_template_simplest_is_actually_trivial(node):
+    # TODO_IR node.trivial is sound but not complete for floats.
+    assume(node.ir_type != "float")
+    data = ConjectureData.for_ir_tree((NodeTemplate("simplest", size=BUFFER_SIZE_IR),))
+    getattr(data, f"draw_{node.ir_type}")(**node.kwargs)
+    assert len(data.ir_nodes) == 1
+    assert data.ir_nodes[0].trivial
