@@ -8,15 +8,18 @@
 # v. 2.0. If a copy of the MPL was not distributed with this file, You can
 # obtain one at https://mozilla.org/MPL/2.0/.
 
+import math
+
 import pytest
 
-from hypothesis import settings
-from hypothesis.internal.compat import int_to_bytes
-from hypothesis.internal.conjecture.data import Status
+from hypothesis import assume, example, given, settings
+from hypothesis.internal.conjecture.data import IRNode, Status
+from hypothesis.internal.conjecture.datatree import compute_max_children
 from hypothesis.internal.conjecture.engine import ConjectureRunner, RunIsComplete
 from hypothesis.internal.entropy import deterministic_PRNG
+from hypothesis.internal.intervalsets import IntervalSet
 
-from tests.conjecture.common import TEST_SETTINGS, buffer_size_limit
+from tests.conjecture.common import TEST_SETTINGS, buffer_size_limit, ir, ir_nodes
 
 
 def test_optimises_to_maximum():
@@ -26,7 +29,7 @@ def test_optimises_to_maximum():
             data.target_observations["m"] = data.draw_integer(0, 2**8 - 1)
 
         runner = ConjectureRunner(test, settings=TEST_SETTINGS)
-        runner.cached_test_function([0])
+        runner.cached_test_function_ir(ir(0))
 
         try:
             runner.optimise_targets()
@@ -49,8 +52,8 @@ def test_optimises_multiple_targets():
             data.target_observations["m + n"] = m + n
 
         runner = ConjectureRunner(test, settings=TEST_SETTINGS)
-        runner.cached_test_function([200, 0])
-        runner.cached_test_function([0, 200])
+        runner.cached_test_function_ir(ir(200, 0))
+        runner.cached_test_function_ir(ir(0, 200))
 
         try:
             runner.optimise_targets()
@@ -71,7 +74,7 @@ def test_optimises_when_last_element_is_empty():
             data.stop_example()
 
         runner = ConjectureRunner(test, settings=TEST_SETTINGS)
-        runner.cached_test_function([250])
+        runner.cached_test_function_ir(ir(250))
 
         try:
             runner.optimise_targets()
@@ -94,7 +97,7 @@ def test_can_optimise_last_with_following_empty():
         runner = ConjectureRunner(
             test, settings=settings(TEST_SETTINGS, max_examples=100)
         )
-        runner.cached_test_function(bytes(101))
+        runner.cached_test_function_ir(ir(0) * 101)
 
         with pytest.raises(RunIsComplete):
             runner.optimise_targets()
@@ -117,7 +120,7 @@ def test_can_find_endpoints_of_a_range(lower, upper, score_up):
         runner = ConjectureRunner(
             test, settings=settings(TEST_SETTINGS, max_examples=1000)
         )
-        runner.cached_test_function(int_to_bytes((lower + upper) // 2, 2))
+        runner.cached_test_function_ir(ir((lower + upper) // 2))
 
         try:
             runner.optimise_targets()
@@ -142,7 +145,7 @@ def test_targeting_can_drive_length_very_high():
         # extend here to ensure we get a valid (non-overrun) test case. The
         # outcome of the test case doesn't really matter as long as we have
         # something for the runner to optimize.
-        runner.cached_test_function(b"", extend=50)
+        runner.cached_test_function_ir([], extend=50)
 
         try:
             runner.optimise_targets()
@@ -163,7 +166,7 @@ def test_optimiser_when_test_grows_buffer_to_invalid():
                 data.mark_invalid()
 
         runner = ConjectureRunner(test, settings=TEST_SETTINGS)
-        runner.cached_test_function(bytes(10))
+        runner.cached_test_function_ir(ir(0) * 10)
 
         try:
             runner.optimise_targets()
@@ -188,7 +191,7 @@ def test_can_patch_up_examples():
                     data.mark_invalid()
 
         runner = ConjectureRunner(test, settings=TEST_SETTINGS)
-        d = runner.cached_test_function([0, 0, 1, 2, 3, 4])
+        d = runner.cached_test_function_ir(ir(0, 0, 1, 2, 3, 4))
         assert d.status == Status.VALID
 
         try:
@@ -211,7 +214,7 @@ def test_optimiser_when_test_grows_buffer_to_overflow():
                     data.mark_invalid()
 
             runner = ConjectureRunner(test, settings=TEST_SETTINGS)
-            runner.cached_test_function(bytes(10))
+            runner.cached_test_function_ir(ir(0) * 10)
 
             try:
                 runner.optimise_targets()
@@ -219,3 +222,58 @@ def test_optimiser_when_test_grows_buffer_to_overflow():
                 pass
 
             assert runner.best_observed_targets["m"] == 100
+
+
+@given(ir_nodes())
+@example(
+    IRNode(
+        ir_type="bytes",
+        value=b"\xb1",
+        kwargs={"min_size": 1, "max_size": 1},
+        was_forced=False,
+    )
+)
+@example(
+    IRNode(
+        ir_type="string",
+        value="aaaa",
+        kwargs={
+            "min_size": 0,
+            "max_size": 10,
+            "intervals": IntervalSet.from_string("abcd"),
+        },
+        was_forced=False,
+    )
+)
+@example(
+    IRNode(
+        ir_type="integer",
+        value=1,
+        kwargs={"min_value": 0, "max_value": 200, "weights": None, "shrink_towards": 0},
+        was_forced=False,
+    )
+)
+def test_optimising_all_nodes(node):
+    assume(compute_max_children(node.ir_type, node.kwargs) > 50)
+    size_function = {
+        "integer": lambda n: n,
+        "float": lambda f: f if math.isfinite(f) else 0,
+        "string": lambda s: len(s),
+        "bytes": lambda b: len(b),
+        "boolean": lambda b: int(b),
+    }
+    with deterministic_PRNG():
+
+        def test(data):
+            v = getattr(data, f"draw_{node.ir_type}")(**node.kwargs)
+            data.target_observations["v"] = size_function[node.ir_type](v)
+
+        runner = ConjectureRunner(
+            test, settings=settings(TEST_SETTINGS, max_examples=50)
+        )
+        runner.cached_test_function_ir([node])
+
+        try:
+            runner.optimise_targets()
+        except RunIsComplete:
+            pass
