@@ -12,20 +12,14 @@ import hypothesis.strategies as st
 from hypothesis.internal.conjecture.data import ConjectureData
 from hypothesis.strategies._internal.strategies import FilteredStrategy
 
-from tests.conjecture.common import run_to_buffer
+from tests.conjecture.common import ir
 
 
 def test_filter_iterations_are_marked_as_discarded():
     variable_equal_to_zero = 0  # non-local references disables filter-rewriting
     x = st.integers().filter(lambda x: x == variable_equal_to_zero)
 
-    @run_to_buffer
-    def buf(data):
-        data.draw_integer(forced=1)
-        data.draw_integer(forced=0)
-        data.mark_interesting()
-
-    data = ConjectureData.for_buffer(buf)
+    data = ConjectureData.for_ir_tree(ir(1, 0))
     assert data.draw(x) == 0
     assert data.has_discards
 
