@@ -21,7 +21,7 @@ class Blocks(SearchStrategy):
         self.n = n
 
     def do_draw(self, data):
-        return data.draw_bytes(self.n)
+        return data.draw_bytes(self.n, self.n)
 
 
 @pytest.mark.parametrize("n", range(1, 5))
@@ -52,12 +52,11 @@ def test_mostly_does_not_duplicate_blocks_even_when_failing(n):
         test()
     except ValueError:
         pass
-    # There are two circumstances in which a duplicate is allowed: We replay
-    # the failing test once to check for flakiness, and then we replay the
-    # fully minimized failing test at the end to display the error. The
-    # complication comes from the fact that these may or may not be the same
-    # test case, so we can see either two test cases each run twice or one
-    # test case which has been run three times.
-    seen_counts = set(counts.values())
-    assert seen_counts in ({1, 2}, {1, 3})
+    # There are three circumstances in which a duplicate is allowed: We replay
+    # the failing test once to check for flakiness, once when shrinking to normalize
+    # to the minimal buffer, and then we replay the fully minimized failing test
+    # at the end to display the error. The complication comes from the fact that
+    # these may or may not be the same test case, so we can see either two test
+    # cases each run twice or one test case which has been run three times.
+    assert set(counts.values()) in ({1, 2, 3}, {1, 4})
     assert len([k for k, v in counts.items() if v > 1]) <= 2
